from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.agents.nothing_agent import NothingAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp,
    check_usage,
    get_pidx_cidx,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_bestest():
    """
    AA + electro A + talent
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "reroll",
            "TEST 1 omni 1",
            "card 0 0 6 7",
            "TEST 2 omni 3",
            "card 2 0 6 7",
            "TEST 3 omni 5",
            "card 2 0 4 7",
            "TEST 4 omni 6",
            "end",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        character:Fischl
        character:Mona
        character:Nahida
        Wine-Stained Tricorne*2
        Timmie*2
        Rana*2
        Strategize*2
        The Bestest Travel Companion!*2
        The Bestest Travel Companion!*20
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                cmd = agent_1.commands[0]
                test_id = get_test_id_from_command(agent_1)
                if test_id == 0:
                    break
                omni_num = int(cmd[-1])
                for color in match.player_tables[1].dice.colors:
                    if color == "OMNI":
                        omni_num -= 1
                assert omni_num == 0
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_changing_shifts():
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 1",
            "card 0 0",
            "card 0 0",
            "TEST 1 status usage 1",
            "skill 0 0 1 2",
            "TEST 1 status 1 usage 1",
            "sw_char 0",
            "TEST 3 status 0",
            "sw_char 1 0",
            "card 0 0",
            "sw_char 0",
            "TEST 4 dice 12",
            "card 0 0",
            "end",
            "TEST 1 status usage 1",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=["sw_card", "choose 1", "sw_char 0 0", "sw_char 1 0", "end", "end"],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Mona
        character:Nahida
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Vanarana
        # Timmie*2
        # Rana*2
        # Covenant of Rock
        # Wind and Freedom
        # The Bestest Travel Companion!*2
        # Changing Shifts*2
        # Strategize*2
        Changing Shifts*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            while True:
                test_id = get_test_id_from_command(agent_0)
                if test_id == 1:
                    assert len(match.player_tables[0].team_status) == 1
                    assert match.player_tables[0].team_status[0].usage == 1
                elif test_id == 3:
                    assert len(match.player_tables[0].team_status) == 0
                elif test_id == 4:
                    assert len(match.player_tables[0].dice.colors) == 12
                else:
                    break
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_toss_up():
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "card 0 0",
            "TEST 2 reroll req 2 time",
            "reroll 0 1 2 3 4 5 6 7",
            "TEST 1 reroll req 1 time",
            "reroll 0 1 2 3 4 5 6 7",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "card 0 0",
            "reroll 0 1 2 3 4 5 6 7",
            "reroll 0 1 2 3 4 5 6 7",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Mona
        character:Nahida
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Vanarana
        # Timmie*2
        # Rana*2
        # Covenant of Rock
        # Wind and Freedom
        # The Bestest Travel Companion!*2
        # Changing Shifts*2
        # Toss-Up
        # Strategize*2
        Toss-Up*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            while True:
                test_id = get_test_id_from_command(agent_0)
                if test_id == 0:
                    break
                assert len(match.requests) == 1
                req = match.requests[0]
                assert req.name == "RerollDiceRequest"
                assert req.reroll_times == test_id
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_i_havent_lost_yet():
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "TEST 1 no card can use",
            "skill 0 0 1 2",
            "choose 1",
            "TEST 2 cards can use",
            "end",
            "TEST 1 card cannot use",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "choose 1",
            "TEST 2 card can use",
            "end",
            "TEST 1 no status cannot use card",
            "end",
            "skill 1 0 1 2",
            "choose 3",
            "card 0 0",
            "TEST 3 dice 14",
            "end",
            "sw_char 4 0",
            "choose 3",
            "skill 1 0 1 2",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "skill 1 0 1 2",
            "end",
            "choose 1",
            "card 2 0",
            "TEST 1 charge 1 status cannot use card",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "end",
            "end",
            "choose 1",
            "skill 1 0 1 2",
            "end",
            "skill 3 0 1 2",
            "choose 3",
            "card 0 0",
            "TEST 2",
            "end",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        # character:Fischl
        # character:Mona
        character:Nahida*10
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Vanarana
        # Timmie*2
        # Rana*2
        # Covenant of Rock
        # Wind and Freedom
        # The Bestest Travel Companion!*2
        # Changing Shifts*2
        # Toss-Up
        # Strategize*2
        # I Haven't Lost Yet!*2
        I Haven't Lost Yet!*30
        """
    )
    for character in deck.characters:
        character.hp = 2
        character.max_hp = 2
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.character_number = 10
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            while True:
                test_id = get_test_id_from_command(agent_0)
                if test_id == 1:
                    assert len(match.player_tables[0].team_status) == 0
                    for req in match.requests:
                        assert req.name != "UseCardRequest"
                elif test_id == 2:
                    use_card_req = 0
                    for req in match.requests:
                        if req.name == "UseCardRequest":
                            use_card_req += 1
                    assert use_card_req > 0
                elif test_id == 3:
                    assert len(match.player_tables[0].dice.colors) == 14
                else:
                    break
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    table = match.player_tables[1]
                    active = table.get_active_character()
                    assert active.charge == 1
                    assert len(table.team_status) == 1
                    assert table.team_status[0].name == "I Haven't Lost Yet!"
                    for req in match.requests:
                        assert req.name != "UseCardRequest"
                elif test_id == 2:
                    table = match.player_tables[1]
                    active = table.get_active_character()
                    assert active.charge == 1
                    assert len(table.team_status) == 2
                    assert table.team_status[1].name == "I Haven't Lost Yet!"
                    for req in match.requests:
                        assert req.name != "UseCardRequest"
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_old_i_havent_lost_yet():
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "skill 0 0 1 2",
            "choose 1",
            "end",
            "skill 1 0 1 2",
            "choose 3",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "skill 1 0 1 2",
            "end",
            "choose 1",
            "card 2 0",
            "TEST 1 charge 0 status",
            "card 0 0",
            "skill 3 0 1 2",
            "end",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        # character:Fischl
        # character:Mona
        character:Nahida*10
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Vanarana
        # Timmie*2
        # Rana*2
        # Covenant of Rock
        # Wind and Freedom
        # The Bestest Travel Companion!*2
        # Changing Shifts*2
        # Toss-Up
        # Strategize*2
        # I Haven't Lost Yet!*2
        """
    )
    old = {"name": "I Haven't Lost Yet!", "version": "3.3"}
    deck_dict = deck.dict()
    deck_dict["cards"] += [old] * 30
    deck = Deck(**deck_dict)
    for character in deck.characters:
        character.hp = 2
        character.max_hp = 2
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.character_number = 10
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    table = match.player_tables[1]
                    active = table.get_active_character()
                    assert active.charge == 1
                    assert len(table.team_status) == 0
                    use_card_req = 0
                    for req in match.requests:
                        if req.name == "UseCardRequest":
                            use_card_req += 1
                    assert use_card_req > 0
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_leave_it_to_me():
    cmd_records = [
        [
            "sw_card",
            "choose 1",
            "card 0 0",
            "card 0 0",
            "TEST 1 status 1",
            "skill 0 0 1 2",
            "TEST 1 status 1",
            "sw_char 0 0",
            "TEST 1 status 1",
            "sw_char 1 0",
            "TEST 2 status 0",
            "sw_char 2 0",
            "TEST 3 opponent ended",
            "end",
        ],
        ["sw_card", "choose 1", "sw_char 0 0", "sw_char 1 0", "end"],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Mona
        character:Nahida
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Vanarana
        # Timmie*2
        # Rana*2
        # Covenant of Rock
        # Wind and Freedom
        # The Bestest Travel Companion!*2
        # Changing Shifts*2
        # Toss-Up
        # Strategize*2
        # I Haven't Lost Yet!*2
        # Leave It to Me!
        Leave It to Me!*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            while True:
                test_id = get_test_id_from_command(agent_0)
                if test_id == 1:
                    table = match.player_tables[0]
                    assert len(table.team_status) == 1
                    assert table.team_status[0].name == "Leave It to Me!"
                elif test_id == 2:
                    table = match.player_tables[0]
                    assert len(table.team_status) == 0
                elif test_id == 3:
                    assert match.player_tables[1].has_round_ended
                else:
                    break
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_claxs_art():
    cmd_records = [
        ["sw_card", "choose 0", "end", "end"],
        [
            "sw_card",
            "choose 0",
            "TEST 1 no card can use",
            "skill 0 0 1 2",
            "TEST 1 no card can use",
            "sw_char 1 0",
            "TEST 3 card can use",
            "card 0 0 0",
            "TEST 2 0 1",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "card 0 0 0",
            "TEST 2 0 0 2",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "end",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "sw_char 3 0",
            "card 0 0 0",
            "TEST 2 0 0 2 2",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "card 0 0 0",
            "TEST 2 2 0 1 1",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        # character:Fischl
        # character:Mona
        character:Nahida*10
        # Gambler's Earrings*2
        # Wine-Stained Tricorne*2
        # Vanarana
        # Timmie*2
        # Rana*2
        # Covenant of Rock
        # Wind and Freedom
        # The Bestest Travel Companion!*2
        # Changing Shifts*2
        # Toss-Up
        # Strategize*2
        # I Haven't Lost Yet!*2
        # Leave It to Me!
        # Calx's Arts*2
        Calx's Arts*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            agent = agent_1
            while True:
                cmd = agent.commands[0]
                test_id = get_test_id_from_command(agent)
                if test_id == 1:
                    for req in match.requests:
                        assert req.name != "UseCardRequest"
                elif test_id == 2:
                    charges = [int(x) for x in cmd.strip().split()[2:]]
                    for c, cc in zip(match.player_tables[1].characters, charges):
                        assert c.charge == cc
                elif test_id == 3:
                    card_req = 0
                    for req in match.requests:
                        if req.name == "UseCardRequest":
                            card_req += 1
                    assert card_req == len(match.player_tables[1].hands)
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_heavy_strike():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 0 0 0",
            "sw_char 1 0",
            "TEST 1 10 7 10 10 10 10",
            "card 0 0 0",
            "skill 0 0 1 2",
            "TEST 1 10 4 10 7 10 10",
            "TEST 2 p1c0 1 status",
            "TEST 2 p0c0 1 status",
            "end",
            "TEST 2 p1c0 0 status",
            "TEST 2 p0c0 0 status",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "card 0 0 0",
            "card 0 0 0",
            "skill 0 0 1 2",
            "TEST 1 10 7 10 7 10 10",
            "card 0 0 0",
            "skill 1 0 1 2",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:PyroMobMage
        character:Arataki Itto
        character:Noelle
        Heavy Strike*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                # test 2 p0c0 1 status
                cmd = cmd.split()
                pnum = int(cmd[2][1])
                cnum = int(cmd[2][3])
                count = int(cmd[3])
                assert len(match.player_tables[pnum].characters[cnum].status) == count
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_heavy_strike_2():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 0 0 1 2",
            "sw_char 1 0",
            "card 0 0 0",
            "skill 0 0 1 2",
            "TEST 1 10 10 10 9 6 9",
            "end",
        ],
        ["sw_card", "choose 1", "end"],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:ElectroMobMage
        character:CryoMobMage
        character:Noelle
        Heavy Strike*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_friendship_eternal():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "reroll",
            "TEST 1 no card can use",
            "tune 0 0",
            "tune 0 0",
            "TEST 2 card can use",
            "card 0 0 0 1",
            "TEST 3 card number 4 5",
            "tune 0 0",
            "tune 0 3",
            "tune 0 4",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "reroll",
            "TEST 2 can use card",
            "card 0 0 0 1",
            "TEST 2 can use card",
            "card 0 0 2 3",
            "TEST 3 card number 4 4",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Rhodeia of Loch
        character:Nahida
        Friendship Eternal*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                for req in match.requests:
                    assert req.name != "UseCardRequest"
            elif test_id == 2:
                found = False
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        found = True
                assert found
            elif test_id == 3:
                cmd = cmd.split()
                numbers = [int(cmd[-2]), int(cmd[-1])]
                for table, num in zip(match.player_tables, numbers):
                    assert len(table.hands) == num
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_unseen_razor():
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "card 0 1 0 1 2",
            "TEST 1 4 card can use",
            "skill 1 0 1 2",
            "card 1 0",
            "TEST 3 p0c2 no equip",
            "card 0 1 0",
            "skill 0 0 1 2",
            "TEST 2 p0 2+2 usage",
            "card 1 0",
            "end",
            "end",
            "end",
        ],
        [
            "sw_card",
            "choose 2",
            "TEST 1 0 card can use",
            "end",
            "TEST 4 p0 2 status",
            "end",
            "card 7 1 0 1 2",
            "card 7 1 0 1 2",
            "skill 1 0 1 2",
            "card 0 0",
            "card 6 1 0",
            "skill 1 0 1 2",
            "TEST 2 p1 2+2 usage",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Barbara
        character:Arataki Itto
        character:Noelle
        Where Is the Unseen Razor?*15
        The Bell*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.split()
                cnum = int(cmd[2])
                counter = 0
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        counter += 1
                assert cnum == counter
            elif test_id == 2:
                cmd = cmd.split()
                pnum = int(cmd[2][1])
                status = match.player_tables[pnum].team_status
                assert len(status) == 2
                for s in status:
                    assert s.usage == 2
            elif test_id == 3:
                assert match.player_tables[0].characters[2].weapon is None
            elif test_id == 4:
                assert len(match.player_tables[0].team_status) == 2
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_send_off_new_and_old():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "TEST 1 no card can use",
            "skill 2 15 14 13 12 11",
            "TEST 1 no card can use",
            "sw_char 1 10",
            "TEST 3 p1 usage 2 3",
            "card 2 1 9 8",
            "TEST 3 p1 usage 2 1",
            "end",
        ],
        [
            "sw_card 2 3 4",
            "choose 1",
            "sw_char 0 15",
            "TEST 2 card can use",
            "card 0 0 14 13",
            "TEST 3 p0 usage 0 2",
            "card 0 1 12 11",
            "TEST 3 p0 usage 0",
            "card 2 0 10 9",
            "TEST 3 p0 usage",
            "skill 2 8 7 6 5 4",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Rhodeia of Loch*3
        Send Off*15
        Send Off@3.3*15
        """
    )
    # use old version cards
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                for req in match.requests:
                    assert req.name != "UseCardRequest"
            elif test_id == 2:
                counter = 0
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        counter += 1
                assert counter > 0
            elif test_id == 3:
                cmd = cmd.split()
                pnum = int(cmd[2][1])
                usage = [int(x) for x in cmd[4:]]
                assert len(match.player_tables[pnum].summons) == len(usage)
                for s, u in zip(match.player_tables[pnum].summons, usage):
                    assert s.usage == u
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_plunge_strike():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "TEST 2 card target 1 2",
            "skill 1 0 1 2",
            "sw_char 1 0",
            "TEST 2 card target 0 2",
            "card 0 0 0 1 2",
            "sw_char 1 0",
            "TEST 1 7 9 10 7 10 8",
            "sw_char 0 0",
            "TEST 1 5 9 10 7 10 8",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "TEST 3 dice number 9 12",
            "TEST 1 7 10 10 7 10 8",
            "sw_char 1 0",
            "card 0 0 0 1 2",
            "skill 0 0 1",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Klee
        character:Arataki Itto
        character:Rhodeia of Loch
        Plunging Strike*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                cnum = [int(x) for x in cmd[4:]]
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert len(req.targets) == len(cnum)
                        for t, c in zip(req.targets, cnum):
                            assert t.character_idx == c  # type: ignore
            elif test_id == 3:
                assert len(match.player_tables[0].dice.colors) == 9
                assert len(match.player_tables[1].dice.colors) == 12
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_star_quick_dream_vennessa_exile():
    cmd_records = [
        [
            "sw_card",
            "choose 1",
            "sw_char 0 15",
            "card 4 0 14 13 12",
            "TEST 2 p0 dice 16",
            "card 0 0",
            "card 1 0 15",
            "TEST 3 card 0 cost 2",
            "card 0 0 13 11",
            "skill 1 12 10 9",
            "TEST 4 card 0 can use",
            "card 0 0 9",
            "TEST 5 p0 summon 4",
            "skill 1 8 7 6",
            "end",
            "TEST 1 5 10 10 9 10 10",
            "TEST 6 p1c1 charge 2",
            "TEST 6 p1c2 charge 3",
            "sw_char 1 15",
            "skill 2 14 13 12 11 10",
            "sw_char 0 9",
            "skill 2 8 7 6 5",
            "TEST 1 5 10 10 1 7 7",
            "TEST 6 p0c1 charge 2",
            "TEST 6 p0c2 charge 3",
            "end",
            "choose 1",
            "end",
            "end",
            "card 6 0 15 14 13",
            "card 0 0 15 12 11",
            "card 5 0",
            "end",
            "card 5 0 15 14 13",
            "end",
            "card 7 0 15",
            "end",
            "card 6 0",
            "card 5 0",
            "card 5 0",
            "card 6 0",
            "end",
            "sw_char 2 15",
            "sw_char 1 14",
            "sw_char 2 13",
            "end",
            "end",
            "end",
            "end",
            "end",
            "end",
            "end",
        ],
        [
            "sw_card 3",
            "choose 1",
            "card 2 0 15 14",
            "card 2 0",
            "card 1 0",
            "TEST 3 card 0 cost 1",
            "TEST 3 card 2 cost 0",
            "card 2 0",
            "sw_char 0 13",
            "skill 1 12 11 10",
            "TEST 5 p0 summon 4",
            "card 1 0 9 8",
            "end",
            "card 0 0 15",
            "skill 2 14 13 12 11",
            "end",
            "choose 1",
            "TEST 7 card 3 cannot use",
            "skill 2 15 14 13 12 11",
            "sw_char 2 10",
            "end",
            "TEST 7 card 1 cannot use",
            "TEST 7 card 3 cannot use",
            "end",
            "end",
            "end",
            "card 5 0 15 14 13",
            "card 0 0 15 12 11",
            "tune 4 15",
            "tune 3 15",
            "tune 1 15",
            "tune 0 12",
            "end",
            "card 2 0 15 14",
            "card 1 1 13 12",
            "sw_char 1 11",
            "end",
            "sw_char 2 15",
            "skill 2 14 13 12 11",
            "sw_char 1 10",
            "end",
            "card 7 1 15 14",
            "end",
            "card 7 0 15 14 13",
            "tune 5 15",
            "tune 2 15",
            "end",
            "end",
            "card 3 0",
            "card 6 0 15 14 13",
            "card 4 1 15",
            "card 6 1 14 13",
            "card 5 0 12 11",
            "end",
            "end",
            "card 5 0",
            "skill 3 15 14 13",
            "card 2 0 12 11",
            "card 1 0 10 9",
            "skill 3 8 7 6",
            "TEST 6 p1c2 charge 1",
            "TEST 6 p1c0 charge 0",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Raiden Shogun
        character:Ganyu@3.3
        character:Keqing
        Quick Knit*6
        Starsigns*6
        The Legend of Vennessa*6
        Rhythm of the Great Dream*6
        Exile's Circlet*6
        Raven Bow*6
        Where Is the Unseen Razor?*6
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    match.config.max_round_number = 999
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                dnum = int(cmd[4])
                assert len(match.player_tables[pidx].dice.colors) == dnum
            elif test_id == 3:
                cmd = cmd.split()
                cidx = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == cidx:
                        assert req.cost.total_dice_cost == cost
            elif test_id == 4:
                found = False
                cmd = cmd.split()
                cidx = int(cmd[3])
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == cidx:
                        found = True
                assert found
            elif test_id == 5:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                usages = [int(x) for x in cmd[4:]]
                summons = match.player_tables[pidx].summons
                assert len(summons) == len(usages)
                for s, u in zip(summons, usages):
                    assert s.usage == u
            elif test_id == 6:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                cidx = int(cmd[2][3])
                charge = int(cmd[4])
                assert match.player_tables[pidx].characters[cidx].charge == charge
            elif test_id == 7:
                cmd = cmd.split()
                cidx = int(cmd[3])
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert req.card_idx != cidx
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_master_weapon_artifact():
    cmd_records = [
        [
            "sw_card 3",
            "choose 0",
            "end",
            "end",
            "card 0 0 15 14 13",
            "skill 1 12 11 10",
            "sw_char 1 10",
            "TEST 1 p1 dice 10",
            "card 2 0",
            "TEST 2 p0c0 no weapon",
            "TEST 3 p0c1 has weapon",
            "skill 1 9 8 7",
            "sw_char 2 6",
            "TEST 1 p1 dice 7",
            "card 0 2 5 4 3",
            "end",
            "card 4 1",
            "TEST 4 p0c1 has artifact",
            "sw_char 1 15",
            "TEST 1 p0 dice 16",
            "card 0 1",
            "sw_char 2 15",
            "TEST 1 p0 dice 15",
            "card 5 0 14 13 12",
            "card 4 0",
            "TEST 5 p0c1 usage",
            "end",
            "card 5 1",
            "sw_char 1 15",
            "card 4 1",
            "sw_char 2 15",
            "TEST 1 p0 dice 16",
            "end",
        ],
        [
            "sw_card 3 2",
            "choose 2",
            "end",
            "end",
            "TEST 1 p0 dice 11",
            "card 3 1 15 14 13",
            "sw_char 1 12",
            "skill 1 11 10 9",
            "TEST 1 p0 dice 7",
            "sw_char 0 9",
            "card 1 0",
            "skill 1 8 7 6",
            "card 0 1 6 5 4",
            "sw_char 2 3",
            "end",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "end",
            "skill 0 15 14 13",
            "TEST 1 p0 dice 15",
            "skill 0 12 11 10",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.1
        character:Fischl
        character:Collei
        character:Nahida
        Master of Weaponry*5
        Blessing of the Divine Relic's Installation*5
        Master of Weaponry@4.0*5
        Blessing of the Divine Relic's Installation@3.3*5
        King's Squire*5
        Sacrificial Bow*5
        Tenacity of the Millelith*5
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].dice.colors) == int(cmd[4])
            elif test_id in [2, 3, 4, 5]:
                cmd = cmd.split()
                pidx, cidx = get_pidx_cidx(cmd)
                character = match.player_tables[pidx].characters[cidx]
                if test_id == 2:
                    assert character.weapon is None
                elif test_id == 3:
                    assert character.weapon is not None
                elif test_id == 4:
                    assert character.artifact is not None
                else:
                    assert test_id == 5
                    assert len(character.status) == 0
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_new_wind_and_freedom_and_crane():
    cmd_records = [
        [
            "sw_card 1 2",
            "choose 0",
            "sw_char 2 15",
            "skill 0 14 13 12",
            "skill 0 11 10 9",
            "sw_char 0 8",
            "card 0 0 7",
            "end",
            "sw_char 1 15",
            "card 0 0 14",
            "card 0 0 13",
            "sw_char 0 12",
            "skill 1 11 10 9",
            "TEST 2 p0 active 0",
            "skill 0 8 7 6",
            "TEST 2 p0 active 1",
            "TEST 1 4 7 7 8 3 10",
            "skill 0 5 4 3",
            "end",
            "skill 2 15 14 13 12",
            "sw_char 0 11",
            "card 4 0 10",
            "card 3 0 9",
            "skill 0 8 7 6",
            "TEST 1 4 7 5 0 2 10",
            "TEST 2 p0 active 2",
            "TEST 3 p0 usage 1",
            "skill 0 5 4 3",
            "end",
            "TEST 4 card all can use",
            "end",
            "choose 2",
            "card 4 0 15",
            "card 3 0",
            "sw_char 1 14",
            "sw_char 2 13",
            "choose 1",
            "skill 0 0 1 2",
            "TEST 3 p0 usage 1",
            "end",
        ],
        [
            "sw_card 1 2",
            "choose 1",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "skill 0 9 8 7",
            "end",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "skill 2 9 8 7",
            "sw_char 0 6",
            "sw_char 1 5",
            "TEST 2 p0 active 2",
            "end",
            "TEST 3 p0 status",
            "sw_char 0 15",
            "skill 0 14 13 12",
            "end",
            "choose 1",
            "choose 2",
            "TEST 5 card cannot use except wind and freedom",
            "end",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "skill 0 9 8 7",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.1
        character:Kaedehara Kazuha
        character:Klee
        character:Kaeya
        When the Crane Returned*10
        Leave It to Me!*10
        Wind and Freedom*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                assert match.player_tables[pidx].active_character_idx == int(cmd[4])
            elif test_id == 3:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                status = match.player_tables[pidx].team_status
                check_usage(status, cmd[4:])
            elif test_id == 4:
                can = []
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        can.append(req.card_idx)
                can.sort()
                assert can == list(range(len(can)))
            elif test_id == 5:
                can = []
                pidx = -1
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        pidx = req.player_idx
                        can.append(req.card_idx)
                hands = match.player_tables[pidx].hands
                for cid, card in enumerate(hands):
                    assert (card.name == "Wind and Freedom") == (cid in can)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_moonpiercer_crown_watatsumi_yayoi_gandharva_pankration():
    cmd_records = [
        [
            "sw_card 4",
            "choose 0",
            "card 2 0 15 14 13",
            "skill 1 12",
            "skill 0 11 10 9",
            "sw_char 1 8",
            "skill 1 7 6 5",
            "end",
            "card 2 1 0",
            "TEST 2 p0c1 usage counter 0 0",
            "skill 1 1 14 13",
            "TEST 1 10 9 10 8 7 10",
            "card 0 0",
            "sw_char 2 1",
            "card 0 0 1",
            "sw_char 1 1",
            "sw_char 2 0",
            "sw_char 1 0",
            "sw_char 2 0",
            "sw_char 1 0",
            "sw_char 2 0",
            "sw_char 1 0",
            "sw_char 2 0",
            "sw_char 1 1",
            "sw_char 0 0",
            "TEST 3 p0 dice OMNI 1",
            "TEST 4 p0 support 2",
            "sw_char 2 0",
            "TEST 5 p0 hand 7",
            "TEST 3 p0 dice 0",
            "TEST 4 p0 support 2",
            "TEST 5 p1 hand 4",
            "end",
            "skill 2 15 14 13 12 11",
            "skill 1 10 9 8",
            "sw_char 1 7",
            "TEST 1 10 8 10 6 8 6",
            "TEST 2 p1c1 usage counter 0 1",
            "skill 0 6 5 4",
            "TEST 2 p1c1 usage counter 2 0",
            "sw_char 0 3",
            "TEST 1 7 8 10 10 10 10",
            "TEST 2 p1c1 usage counter 0 0",
            "sw_char 1 2",
            "sw_char 2 1",
            "sw_char 0 0",
            "TEST 3 p0 dice 1",
            "sw_char 1 0",
            "end",
            "TEST 2 p0c1 usage counter 1 0",
            "skill 2 15 14 13",
            "skill 1 12 11 10",
            "TEST 6 card 3 can use",
            "card 4 0",
            "sw_char 0 9",
            "end",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "sw_char 1 9",
            "sw_char 0 8",
            "TEST 6 card 3 can use",
            "sw_char 1 7",
            "TEST 7 card 3 cannot use",
            "end",
        ],
        [
            "sw_card 2 4",
            "choose 1",
            "card 0 1 0",
            "skill 1 0 14 13",
            "sw_char 0 11",
            "skill 0 1 10 9",
            "end",
            "TEST 1 10 10 10 8 8 10",
            "TEST 2 p1c1 usage counter 0 2",
            "sw_char 1 15",
            "skill 1 14 13 12",
            "card 5 0",
            "card 4 0 11",
            "TEST 8 card 0 cost 3",
            "sw_char 2 10",
            "TEST 4 p0 support 3",
            "sw_char 1 9",
            "sw_char 2 8",
            "sw_char 1 7",
            "sw_char 2 6",
            "sw_char 1 5",
            "sw_char 2 4",
            "sw_char 1 3",
            "sw_char 2 2",
            "sw_char 1 1",
            "sw_char 2 0",
            "TEST 5 p0 hand 3",
            "end",
            "TEST 2 p1c1 usage count 1 1",
            "sw_char 0 15",
            "sw_char 2 14",
            "sw_char 1 13",
            "skill 0 12 11 10",
            "TEST 6 card 1 can use",
            "card 1 0",
            "skill 2 9 8 7",
            "skill 0 6 5 4",
            "sw_char 0 3",
            "sw_char 1 2",
            "TEST 3 p0 dice 0",
            "sw_char 0 1",
            "sw_char 1 0",
            "TEST 5 p1 hand 7",
            "card 4 2",
            "end",
            "TEST 2 p0c1 usage counter 2 0",
            "skill 1 15 14 13",
            "TEST 1 10 9 10 10 7 10",
            "sw_char 0 12",
            "TEST 6 card 7 can use",
            "skill 0 11 10 9",
            "TEST 5 p1 hand 10",
            "TEST 7 card 7 cannot use",
            "end",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "sw_char 1 9",
            "sw_char 0 8",
            "TEST 6 card 7 can use",
            "sw_char 1 7",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.1
        character:Yaoyao
        character:Barbara
        character:Nahida
        Moonpiercer*5
        Crown of Watatsumi*5
        Yayoi Nanatsuki*5
        Gandharva Ville*5
        Pankration!*5
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx, cidx = get_pidx_cidx(cmd)
                usage = int(cmd[5])
                counter = int(cmd[6])
                artifact = match.player_tables[pidx].characters[cidx].artifact
                assert artifact is not None
                assert artifact.usage == usage
                assert artifact.counter == counter  # type: ignore
            elif test_id == 3:
                pidx = int(cmd[2][1])
                tot = int(cmd[-1])
                colors = match.player_tables[pidx].dice.colors
                assert len(colors) == tot
                if "OMNI" in " ".join(cmd):
                    assert colors[0] == "OMNI"
            elif test_id == 4:
                pidx = int(cmd[2][1])
                support = match.player_tables[pidx].supports
                check_usage(support, cmd[4:])
            elif test_id == 5:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].hands) == int(cmd[4])
            elif test_id == 6:
                found = False
                cidx = int(cmd[3])
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == cidx:
                        found = True
                assert found
            elif test_id == 7:
                cidx = int(cmd[3])
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert req.card_idx != cidx
            elif test_id == 8:
                cidx = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == cidx:
                        assert req.cost.total_dice_cost == cost
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_lyresong():
    """
    When not sonsume, Lyresong status remains. No need to write test command.
    """
    cmd_records = [
        ["sw_card 1 2", "choose 0", "end"],
        [
            "sw_card 2 3",
            "choose 0",
            "card 1 1 15",
            "card 2 0",
            "card 1 0",
            "card 0 0",
            "card 1 1",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.2
        character:Keqing
        character:Nahida
        character:Ganyu
        Gambler's Earrings*10
        Lyresong*10
        Rhythm of the Great Dream*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_i_havent_lost_yet_in_deck():
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "skill 0 9 8",
            "skill 2 7 6 5",
            "end",
            "end",
            "sw_char 0 15",
            "sw_char 2 14",
            "skill 0 13 12",
            "TEST 2 not all can use",
            "end",
        ],
        [
            "sw_card",
            "choose 2",
            "sw_char 1 15",
            "skill 1 14 13 12",
            "end",
            "end",
            "sw_char 2 15",
            "sw_char 0 14",
            "sw_char 1 13",
            "choose 0",
            "TEST 1 all card can use",
            "card 7 0 12",
            "TEST 1 all card can use",
            "card 7 0 11",
            "card 8 0 10",
            "card 9 0 9",
            "TEST 1 all card can use",
            "card 1 0",
            "card 0 0",
            "card 0 0",
            "card 0 0",
            "card 0 0",
            "end",
            "TEST 2 not all can use",
            "card 3 0 15",
            "card 4 0 14",
            "TEST 2 not all can use",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.2
        character:Sangonomiya Kokomi
        character:Fischl
        character:Arataki Itto
        I Haven't Lost Yet!@3.3*15
        Strategize*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cards = []
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        cards.append(req.card_idx)
                cards.sort()
                assert cards == list(range(len(cards)))
            elif test_id == 2:
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert (
                            match.player_tables[req.player_idx].hands[req.card_idx].name
                            != "I Haven't Lost Yet!"
                        )
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_unseen_razor_2():
    cmd_records = [
        [
            "sw_card",
            "choose 1",
            "card 1 0 15 14 13",
            "card 1 0",
            "sw_char 2 12",
            "card 1 0 11",
            "card 1 0",
            "end",
            "card 2 1 15 14 13",
        ],
        [
            "sw_card 3 2 1",
            "choose 2",
            "card 2 0 15 14 13",
            "card 1 0",
            "card 1 1 12",
            "sw_char 1 11",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Barbara
        character:Arataki Itto
        character:Noelle
        Where Is the Unseen Razor?*15
        The Bell*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_i_havent_lost_yet_after_fix():
    cmd_records = [
        [
            "sw_card 1",
            "choose 1",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "skill 1 9 8 7",
            "end",
            "card 4 0",
            "skill 0 15 14",
            "choose 0",
            "TEST 1 all card can use",
            "end",
            "TEST 2 no card can use",
            "skill 1 15 14 13",
        ],
        [
            "sw_card",
            "choose 1",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "skill 1 9 8 7",
            "end",
            "choose 2",
            "skill 0 15 14 13",
            "card 0 0",
            "TEST 1 all card can use",
            "end",
            "TEST 2 no card can use",
            "skill 1 15 14 13",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.3
        character:Kaedehara Kazuha
        character:Klee
        character:Kaeya
        I Haven't Lost Yet!*10
        Liben*1
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cards = []
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        cards.append(req.card_idx)
                cards.sort()
                assert cards == list(range(len(cards)))
            elif test_id == 2:
                for req in match.requests:
                    assert req.name != "UseCardRequest"
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    # test_bestest()
    # test_changing_shifts()
    # test_toss_up()
    # test_old_i_havent_lost_yet()
    # test_leave_it_to_me()
    # test_claxs_art()
    # test_heavy_strike_2()
    # test_friendship_eternal()
    # test_unseen_razor()
    # test_send_off_new_and_old()
    # test_plunge_strike()
    # test_star_quick_dream_vennessa_exile()
    # test_master_weapon_artifact()
    # test_new_wind_and_freedom_and_crane()
    # test_moonpiercer_crown_watatsumi_yayoi_gandharva_pankration()
    # test_lyresong()
    # test_i_havent_lost_yet_in_deck()
    # test_unseen_razor_2()
    test_i_havent_lost_yet_after_fix()



from typing import Any, Dict, List, Literal

from ....status.team_status.event_cards import (
    StoneAndContracts_3_7 as StoneAndContractsStatus_3_7,
)

from .....utils.class_registry import register_class
from ....action import CreateDiceAction, DrawCardAction, RemoveObjectAction
from .....utils.desc_registry import DescDictType
from ....card.event.resonance import StoneAndContracts_3_7


class StoneAndContractsStatus_4_3(StoneAndContractsStatus_3_7):
    version: Literal["4.3"] = "4.3"

    def event_handler_ROUND_PREPARE(
        self, event: Any, match: Any
    ) -> List[CreateDiceAction | RemoveObjectAction | DrawCardAction]:
        """
        When round prepare, create 3 omni element and remove self.
        """
        return super().event_handler_ROUND_PREPARE(event, match) + [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=1,
                draw_if_filtered_not_enough=True,
            )
        ]


class StoneAndContracts_4_3(StoneAndContracts_3_7):
    version: Literal["4.3"] = "4.3"


desc: Dict[str, DescDictType] = {
    "TEAM_STATUS/Stone and Contracts": {
        "descs": {
            "4.3": {
                "zh-CN": "下回合行动阶段开始时：生成3点万能元素，抓一张牌。",
                "en-US": "When the Action Phase of the next Round begins: Create 3 Omni Element, draw a card.",  # noqa: E501
            }
        },
    },
    "CARD/Stone and Contracts": {
        "descs": {
            "4.3": {
                "zh-CN": "下回合行动阶段开始时：生成3点万能元素，抓一张牌。（牌组包含至少2个「璃月」角色，才能加入牌组）",  # noqa: E501
                "en-US": "When the Action Phase of the next Round begins: Create 3 Omni Element, draw a card. (You must have at least 2 Liyue characters in your deck to add this card to your deck.)",  # noqa: E501
            }
        },
    },
}


register_class(StoneAndContractsStatus_4_3 | StoneAndContracts_4_3, desc)



import pydantic
import os
import importlib
from typing import List, Literal, get_origin, get_type_hints
from .class_registry import (  # noqa: F401
    register_class,
    get_instance,
    get_class_list_by_base_class,
)
from .desc_registry import (  # noqa: F401
    DescDictType,
    ExpectedLanguageType,
    update_desc,
    desc_exist,
    get_desc_patch,
)
from .deck_code import (  # noqa: F401
    deck_code_to_deck_str,
    deck_str_to_deck_code,
)


class BaseModel(pydantic.BaseModel):
    class Config:
        extra = pydantic.Extra.forbid  # default forbid extra fields


def list_unique_range_right(data: List[int], minn: int, maxn: int) -> bool:
    """
    Check if the list is unique and in range [minn, maxn).
    """
    if len(data) != len(set(data)):
        return False
    for i in data:
        if i < minn or i >= maxn:
            return False
    return True


def import_all_modules(py_file: str, name: str, exceptions: set[str] = set()) -> None:
    """
    This function is used to activate all modules in the directory for __init__
    and they can do default actions, i.e. register themselves to the class
    registry. To expose inner modules and classes, cannot use this function,
    import manually instead.
    """
    for file in os.listdir(os.path.dirname(py_file)):
        if file[0] == "_" or file[0] == ".":
            continue
        if file.endswith(".py"):
            file = file[:-3]
        if file not in exceptions:  # pragma: no branch
            importlib.import_module("." + file, package=name)


def accept_same_or_higher_version(cls, v: str, values):
    msg = "version annotation must be Literal with one str"
    if not isinstance(v, str):
        raise ValueError(msg)
    version_hints = get_type_hints(cls)["version"]
    if get_origin(version_hints) != Literal:
        raise ValueError(msg)
    version_hints = version_hints.__args__
    if len(version_hints) > 1:
        raise ValueError(msg)
    version_hint = version_hints[0]
    if values["strict_version_validation"] and v != version_hint:
        raise ValueError(f"version {v} is not equal to {version_hint}")
    if v < version_hint:
        raise ValueError(f"version {v} is lower than {version_hint}")
    return version_hint


if __name__ == "__main__":
    pass



import logging
from typing import Any, List

from .consts import ObjectPositionType
from .event import EventArguments
from .action import Actions
from .struct import ObjectPosition
from ..utils import BaseModel


class EventFrame(BaseModel):
    """
    Event frame is a frame of events, contains events with their arguments, the
    object lists that will be triggered by the event arguments, and the
    action lists that has triggered.

    When one action is done, it will generate a new event frame, with various
    number of events, and append it to the event frame list. When there are
    events and no processing objects and actions, the first event will
    trigger objects. When there are triggered objects, the first object
    position will be popped and trigger actions. When there are triggered
    actions, the first action will be popped and trigger events, then add
    a new event frame. If an event frame has no events, no triggered objects
    and no triggered actions, it will be removed from the list.
    """

    events: List[EventArguments]
    processing_event: EventArguments | None = None
    triggered_objects: List[ObjectPosition] = []
    triggered_actions: List[Actions] = []


class EventController(BaseModel):
    frame_list: List[EventFrame] = []

    def has_event(self) -> bool:
        return len(self.frame_list) != 0

    def append(self, event_frame: EventFrame) -> None:
        self.frame_list.append(event_frame)

    def pop(self) -> EventFrame:
        """
        pop and return last event frame.
        """
        return self.frame_list.pop()

    def run_event_frame(self, match: Any) -> None:
        """
        run the event frame. if last event frame contains actions, do actions and
        return; otherwise,
        trigger next object; otherwise, trigger next event; otherwise, pop last frame;
        otherwise, clear trashbin and done triggering.
        When any actions has done, or no valid event frame exist, return.
        """
        while len(self.frame_list):
            event_frame = self.frame_list[-1]
            if len(event_frame.triggered_actions):
                self.act_action(event_frame, match)
                return
            elif len(event_frame.triggered_objects):
                self.get_action(event_frame, match)
            elif len(event_frame.events):
                self.trigger_event(event_frame, match)
            else:
                self.pop()
        # event frame cleared, clear trashbin
        match.trashbin.clear()

    def act_action(self, frame: EventFrame, match: Any) -> None:
        """
        Act the actions. If these actions are triggered by one object (i.e.,
        event_frame.processing_event is not None), do all actions immediately;
        otherwise, actions are generated by system, and only do one action.
        """
        if frame.processing_event is None:
            # do one action
            activated_action = frame.triggered_actions.pop(0)
            logging.info(f"Action activated: {activated_action}")
            event_args = match._act(activated_action)
            match.record_last_action_history()
            self.stack_events(event_args)
        else:
            # do all actions
            event_args = []
            for activated_action in frame.triggered_actions:
                logging.info(f"Action activated: {activated_action}")
                event_args += match._act(activated_action)
                match.record_last_action_history()
            frame.triggered_actions = []
            self.stack_events(event_args)

    def get_action(self, frame: EventFrame, match: Any) -> None:
        """
        Get actions from triggered objects.
        """
        event_arg = frame.processing_event
        assert event_arg is not None
        object_position = frame.triggered_objects.pop(0)
        obj = match.get_object(object_position, event_arg.type)
        handler_name = f"event_handler_{event_arg.type.name}"
        func = getattr(obj, handler_name, None)
        if func is not None:
            frame.triggered_actions = func(event_arg, match)
        self.frame_list[-1] = frame

    def trigger_event(self, event_frame: EventFrame, match: Any) -> None:
        """
        trigger new event to update triggered object lists of a EventFrame.
        it will take first event from events, put it into processing_event,
        and update triggered object lists.
        """
        event_arg = event_frame.events.pop(0)
        event_frame.processing_event = event_arg
        object_list = match.get_object_list()
        # add object in trashbin to list
        for obj in match.trashbin:
            if event_arg.type in obj.available_handler_in_trashbin:
                object_list.append(obj)
        handler_name = f"event_handler_{event_arg.type.name}"
        for obj in object_list:
            # for deck objects, check availability
            if obj.position.area == ObjectPositionType.DECK:
                if event_arg.type not in obj.available_handler_in_deck:
                    continue
            func = getattr(obj, handler_name, None)
            if func is not None:
                event_frame.triggered_objects.append(obj.position)

    def stack_event(self, event_arg: EventArguments) -> EventFrame:
        """
        stack a new event. It will wrap it into a list and call
        self.stack_events.
        """
        return self.stack_events([event_arg])

    def stack_actions(self, actions: List[Actions]) -> EventFrame:
        """
        stack new actions, these actions are not triggered by any object, so when any
        action in it is triggered, its event will be triggered immediately.
        """
        event_frame = EventFrame(events=[], triggered_actions=actions)
        self.frame_list.append(event_frame)
        return event_frame

    def stack_events(self, event_args: List[EventArguments]) -> EventFrame:
        """
        stack events. It will create a new EventFrame with events and
        append it into self.event_frames. Then it will return the event frame.
        """
        frame = EventFrame(
            events=event_args,
        )
        self.frame_list.append(frame)
        return frame



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ..base import StatusBase

from ...action import Actions, CreateObjectAction

from ...event import RoundPrepareEventArguments, SkillEndEventArguments

from ...consts import (
    DamageElementalType,
    DieColor,
    IconType,
    ObjectPositionType,
    SkillType,
)
from ...modifiable_values import (
    CostValue,
    DamageDecreaseValue,
    DamageElementEnhanceValue,
    DamageIncreaseValue,
)

from .base import (
    DefendCharacterStatus,
    ElementalInfusionCharacterStatus,
    RoundCharacterStatus,
    UsageCharacterStatus,
)


class SweepingTime_3_3(RoundCharacterStatus, ElementalInfusionCharacterStatus):
    name: Literal["Sweeping Time"] = "Sweeping Time"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    cost_decrease_usage: int = 1
    infused_elemental_type: DamageElementalType = DamageElementalType.GEO
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def renew(self, new_status: StatusBase) -> None:
        super().renew(new_status)
        # when renew, also renew cost decrease usage
        self.cost_decrease_usage = 1

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        # reset cost decrease usage
        self.cost_decrease_usage = 1
        return super().event_handler_ROUND_PREPARE(event, match)

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        If self use normal attack, and not decreased this round, and has
        cost, then decrease one geo die cost.
        """
        if self.cost_decrease_usage <= 0:
            # out of usage, not modify
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not this character use skill, not modify
            return value
        skill = match.get_object(value.position)
        skill_type: SkillType = skill.skill_type
        if skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack, not modify
            return value
        # decrease geo cost
        if value.cost.decrease_cost(DieColor.GEO):  # pragma: no branch
            # decrease success
            if mode == "REAL":
                self.cost_decrease_usage -= 1
        return value

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self,
        value: DamageElementEnhanceValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageElementEnhanceValue:
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this character use normal attack, not modify
            return value
        return super().value_modifier_DAMAGE_ELEMENT_ENHANCE(value, match, mode)

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this character use normal attack, not modify
            return value
        # modify damage
        value.damage += 2
        return value


class RagingOniKing_4_2(RoundCharacterStatus, ElementalInfusionCharacterStatus):
    name: Literal["Raging Oni King"] = "Raging Oni King"
    version: Literal["4.2"] = "4.2"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS
    damage_increase: int = 1

    infused_elemental_type: DamageElementalType = DamageElementalType.GEO

    status_increase_usage: int = 1

    def __init__(self, *argv, **kwargs) -> None:
        super().__init__(*argv, **kwargs)

    def renew(self, new_status: StatusBase) -> None:
        super().renew(new_status)
        # when renew, also renew status increase usage
        self.status_increase_usage = 1

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        # reset status increase usage
        self.status_increase_usage = 1
        return super().event_handler_ROUND_PREPARE(event, match)

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self,
        value: DamageElementEnhanceValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageElementEnhanceValue:
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this character use normal attack, not modify
            return value
        return super().value_modifier_DAMAGE_ELEMENT_ENHANCE(value, match, mode)

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this character use normal attack, not modify
            return value
        # modify damage
        value.damage += self.damage_increase
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If self use normal attack, and has status increase usage,
        gain superlative superstrength.
        """
        action = event.action
        if action.skill_type != SkillType.NORMAL_ATTACK:
            # not using normal attack
            return []
        if self.position.not_satisfy(
            "both pidx=same cidx=same",
            action.position,
        ):
            # not attack by self
            return []
        if self.status_increase_usage <= 0:
            # out of usage, not trigger
            return []
        # trigger
        self.status_increase_usage -= 1
        position = self.position.set_area(ObjectPositionType.CHARACTER_STATUS)
        return [
            CreateObjectAction(
                object_position=position,
                object_name="Superlative Superstrength",
                object_arguments={},
            )
        ]


class SuperlativeSuperstrength_3_6(UsageCharacterStatus):
    name: Literal["Superlative Superstrength"] = "Superlative Superstrength"
    version: Literal["3.6"] = "3.6"
    usage: int = 1
    max_usage: int = 3
    icon_type: Literal[IconType.ATK_UP_ROCK] = IconType.ATK_UP_ROCK

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this character use normal attack, not modify
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charged attack, not modify
            return value
        # modify damage
        assert self.usage > 0
        self.usage -= 1
        value.damage += 1
        return value

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        If self use charged attack, and has 2 or more usage, decrease one
        unaligned die cost.
        """
        if self.usage < 2:
            # not enough usage, not modify
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not this character use skill, not modify
            return value
        skill = match.get_object(value.position)
        skill_type: SkillType = skill.skill_type
        if skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack, not modify
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charged attack, not modify
            return value
        # try decrease any cost
        value.cost.decrease_cost(None)
        return value


class Stonehide_3_3(ElementalInfusionCharacterStatus, DefendCharacterStatus):
    """
    Combined Stonehide and Stone Force into one.
    """

    name: Literal["Stonehide"] = "Stonehide"
    version: Literal["3.3"] = "3.3"
    usage: int = 3
    max_usage: int = 3

    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1
    decrease_usage_by_damage: bool = False

    infused_elemental_type: DamageElementalType = DamageElementalType.GEO
    icon_type: Literal[IconType.BARRIER] = IconType.BARRIER

    damage_increase_usage: int = 1
    damage_increase_usage_max: int = 1

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        Renew usage
        """
        self.damage_increase_usage = self.damage_increase_usage_max
        return []

    def value_modifier_DAMAGE_DECREASE(
        self, value: DamageDecreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageDecreaseValue:
        value = super().value_modifier_DAMAGE_DECREASE(value, match, mode)
        # check if need to decrease usage 1 more
        if not value.is_corresponding_character_receive_damage(self.position, match):
            # not corresponding character, return
            return value
        if value.damage_elemental_type != DamageElementalType.GEO:
            # not geo damage, return
            return value
        if self.usage > 0:
            # decrease usage
            self.usage -= 1
        return value

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        Increase damage by 1
        """
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not corresponding character, return
            return value
        if self.damage_increase_usage <= 0:
            # no usage, return
            return value
        # increase damage
        value.damage += 1
        self.damage_increase_usage -= 1
        return value


class Petrification_3_7(RoundCharacterStatus):
    name: Literal["Petrification"] = "Petrification"
    version: Literal["3.7"] = "3.7"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS


class RagingOniKing_3_6(RagingOniKing_4_2):
    version: Literal["3.6"]
    damage_increase: int = 2


register_class(
    SweepingTime_3_3
    | RagingOniKing_4_2
    | SuperlativeSuperstrength_3_6
    | Stonehide_3_3
    | Petrification_3_7
    | RagingOniKing_3_6
)



from typing import Dict, Literal

from lpsim.server.card.support.locations import KnightsOfFavoniusLibrary_3_3
from lpsim.server.struct import Cost
from lpsim.utils.class_registry import register_class
from lpsim.utils.desc_registry import DescDictType


class KnightsOfFavoniusLibrary_4_5(KnightsOfFavoniusLibrary_3_3):
    version: Literal["4.5"] = "4.5"
    cost: Cost = Cost()


desc: Dict[str, DescDictType] = {
    "SUPPORT/Knights of Favonius Library": {
        "descs": {
            "4.5": {
                "zh-CN": "$SUPPORT/Knights of Favonius Library|descs|3.3|zh-CN",
                "en-US": "$SUPPORT/Knights of Favonius Library|descs|3.3|en-US",
            }
        },
    },
}


register_class(KnightsOfFavoniusLibrary_4_5, desc)



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....event import ChooseCharacterEventArguments, SwitchCharacterEventArguments

from ....action import CreateObjectAction
from ....modifiable_values import CostValue, DamageIncreaseValue
from ....consts import CostLabels, ObjectPositionType, SkillType
from ....struct import Cost
from .base import RoundEffectArtifactBase


class SkillCostDecreaseArtifact(RoundEffectArtifactBase):
    name: str
    version: str
    cost: Cost
    skill_label: int  # CostLabels

    max_usage_per_round: int = 1

    def value_modifier_COST(
        self,
        value: CostValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> CostValue:
        if self.usage > 0:
            # has usage
            if not self._check_value_self_skill_or_talent(
                value, match, self.skill_label
            ):
                return value
            # can decrease cost
            if (  # pragma: no branch
                value.cost.decrease_cost(value.cost.elemental_dice_color)
            ):
                # decrease cost success
                if mode == "REAL":
                    self.usage -= 1
        return value


class ThunderingPoise_4_0(SkillCostDecreaseArtifact):
    name: Literal["Thundering Poise"]
    version: Literal["4.0"] = "4.0"
    cost: Cost = Cost(any_dice_number=2)
    skill_label: int = CostLabels.NORMAL_ATTACK.value


class VermillionHereafter_4_0(ThunderingPoise_4_0):
    name: Literal["Vermillion Hereafter"]
    cost: Cost = Cost(any_dice_number=3)

    def _attach_status(
        self, player_idx: int, character_idx: int
    ) -> List[CreateObjectAction]:
        """
        attach status
        """
        if (
            player_idx == self.position.player_idx
            and character_idx == self.position.character_idx
            and self.position.area == ObjectPositionType.CHARACTER
        ):
            # equipped and switch to this character
            return [
                CreateObjectAction(
                    object_position=self.position.set_area(
                        ObjectPositionType.CHARACTER_STATUS
                    ),
                    object_name=self.name,
                    object_arguments={},
                )
            ]
        return []

    def event_handler_SWITCH_CHARACTER(
        self, event: SwitchCharacterEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If switch to this character, attach status.
        """
        return self._attach_status(event.action.player_idx, event.action.character_idx)

    def event_handler_CHOOSE_CHARACTER(
        self, event: ChooseCharacterEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If choose this character, attach status.
        """
        return self._attach_status(event.action.player_idx, event.action.character_idx)


class CapriciousVisage_4_0(SkillCostDecreaseArtifact):
    name: Literal["Capricious Visage"]
    version: Literal["4.0"] = "4.0"
    cost: Cost = Cost(any_dice_number=2)
    skill_label: int = CostLabels.ELEMENTAL_SKILL.value


class ShimenawasReminiscence_4_0(CapriciousVisage_4_0):
    name: Literal["Shimenawa's Reminiscence"]
    cost: Cost = Cost(any_dice_number=3)

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        """
        if this character has at least 2 energy and use normal attack or
        elemental skill, +1 damage.
        """
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not current character using skill
            return value
        skill = match.get_object(value.position)
        if skill.skill_type not in [SkillType.NORMAL_ATTACK, SkillType.ELEMENTAL_SKILL]:
            # skill type not match
            return value
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.charge < 2:
            # not enough energy
            return value
        # increase damage
        assert mode == "REAL"
        value.damage += 1
        return value


class ThunderingPoise_3_7(ThunderingPoise_4_0):
    version: Literal["3.7"]
    cost: Cost = Cost(same_dice_number=2)


class VermillionHereafter_3_7(VermillionHereafter_4_0):
    version: Literal["3.7"]
    cost: Cost = Cost(same_dice_number=3)


class CapriciousVisage_3_7(CapriciousVisage_4_0):
    version: Literal["3.7"]
    cost: Cost = Cost(same_dice_number=2)


class ShimenawasReminiscence_3_7(ShimenawasReminiscence_4_0):
    version: Literal["3.7"]
    cost: Cost = Cost(same_dice_number=3)


register_class(
    ThunderingPoise_4_0
    | VermillionHereafter_4_0
    | CapriciousVisage_4_0
    | ShimenawasReminiscence_4_0
    | ThunderingPoise_3_7
    | VermillionHereafter_3_7
    | CapriciousVisage_3_7
    | ShimenawasReminiscence_3_7
)



import pytest
from lpsim.utils import deck_code_to_deck_str, deck_str_to_deck_code
from lpsim import Deck


def test_deck_code():
    deck_str = """
    # nahida mona fischl, with 4.0 wind and freedom
    default_version:4.1
    character:Fischl@3.3
    character:Mona
    character:Nahida
    Gambler's Earrings*2
    Wine-Stained Tricorne*2
    Vanarana
    Timmie*2
    Rana*2
    Covenant of Rock
    Wind and Freedom@4.0
    The Bestest Travel Companion!*2
    Changing Shifts*2
    Toss-Up
    Strategize*2
    I Haven't Lost Yet!*2
    Leave It to Me!
    Calx's Arts*2
    Adeptus' Temptation*2
    Lotus Flower Crisp*2
    Mondstadt Hash Brown*2
    Tandoori Roast Chicken
    """
    deck_str_over_1 = deck_str + "character:Beidou"
    deck_str_over_2 = deck_str + "Strategize"
    deck_str_miss_1 = deck_str.replace("character:Mona", "")
    deck_str_miss_2 = deck_str.replace("Tanodoori Roast Chicken", "")
    deck_str_over_3 = deck_str.replace("character:Mona", "character:Mona*2")
    # check deck str to deck code
    deck_code = deck_str_to_deck_code(deck_str)
    with pytest.raises(ValueError):
        deck_str_to_deck_code(deck_str_over_1)
    with pytest.raises(ValueError):
        deck_str_to_deck_code(deck_str_over_2)
    with pytest.raises(ValueError):
        deck_str_to_deck_code(deck_str_over_3)
    deck_code_miss_1 = deck_str_to_deck_code(deck_str_miss_1)
    deck_code_miss_2 = deck_str_to_deck_code(deck_str_miss_2)

    # if retry time is small, will fail.
    with pytest.raises(ValueError):
        for i in range(10000):  # pragma: no branch
            # it is nearly impossible to generate a legal deck code 10000 times
            deck_str_to_deck_code(deck_str, max_retry_time=1)

    # check deck code to deck str is same
    def check_deck_str_same(str1: str, str2: str) -> bool:
        """
        Will use str to generate deck, and parse names, and sort to confirm
        whether they are same.
        """
        deck1 = Deck.from_str(str1)
        deck2 = Deck.from_str(str2)
        ch_names1 = sorted([x.name for x in deck1.characters])
        ca_names1 = sorted([x.name for x in deck1.cards])
        ch_names2 = sorted([x.name for x in deck2.characters])
        ca_names2 = sorted([x.name for x in deck2.cards])
        return ch_names1 == ch_names2 and ca_names1 == ca_names2

    for _ in range(100):
        deck_str_rev = deck_code_to_deck_str(deck_code)
        deck_str_miss_1_rev = deck_code_to_deck_str(deck_code_miss_1)
        deck_str_miss_2_rev = deck_code_to_deck_str(deck_code_miss_2)
        # revert result should same
        assert check_deck_str_same(deck_str, deck_str_rev)
        assert check_deck_str_same(deck_str_miss_1, deck_str_miss_1_rev)
        assert check_deck_str_same(deck_str_miss_2, deck_str_miss_2_rev)
        # generate new deck code
        deck_code = deck_str_to_deck_code(deck_str)
        deck_code_miss_1 = deck_str_to_deck_code(deck_str_miss_1)
        deck_code_miss_2 = deck_str_to_deck_code(deck_str_miss_2)

    # when version hint is added, output should contain version
    deck_str_rev_version = deck_code_to_deck_str(deck_code, version="4.0")
    assert "default_version:4.0" in deck_str_rev_version

    deck_instance = Deck.from_deck_code(deck_code)
    deck_code_rev = deck_instance.to_deck_code()
    assert check_deck_str_same(deck_str, deck_code_to_deck_str(deck_code_rev))


def test_deck_doce_shuffle():
    deck_str = """
        character:Yelan
        character:Chongyun
        character:Kamisato Ayaka
        Sacrificial Bow
        King's Squire
        Light of Foliar Incision
        Light of Foliar Incision
        Gambler's Earrings
        Gambler's Earrings
        Wagner
        Wagner
        Chef Mao
        Chef Mao
        Liben
        Liben
        Dunyarzad
        Dunyarzad
        Master Zhang
        Mamere
        Mamere
        Elemental Resonance: Woven Ice
        Elemental Resonance: Woven Ice
        Elemental Resonance: Shattering Ice
        Elemental Resonance: Shattering Ice
        Stone and Contracts
        Leave It to Me!
        Heavy Strike
        The Legend of Vennessa
        The Legend of Vennessa
        Northern Smoked Chicken
        Northern Smoked Chicken
        Minty Meat Rolls
        Minty Meat Rolls
    """
    deck_str = "\n".join([x.strip() for x in deck_str.strip().split("\n")])
    # deck_code = 'tqYl1KmzpYX2a7GxsjWVbiixtvWlPGu0rgUWdqizsqa23LG3s5am3KiysZX1JHi1tsWl'  # noqa: E501
    # deck_str = deck_code_to_deck_str(deck_code)
    deck_str_sorted = sorted(deck_str.strip().split("\n"))
    for _ in range(100):  # pragma: no branch
        deck_code_2 = deck_str_to_deck_code(deck_str)
        deck_str_2 = deck_code_to_deck_str(deck_code_2)
        deck_str_2_sorted = sorted(deck_str_2.strip().split("\n"))
        assert deck_str_sorted == deck_str_2_sorted
        if deck_str != deck_str_2:  # pragma: no branch
            break
    else:
        raise AssertionError(
            "this deck code should shuffle to generate valid code, "
            "but with 100 times no different code is generated!"
        )


if __name__ == "__main__":
    test_deck_code()
    test_deck_doce_shuffle()



from typing import List, Literal

from ....utils.class_registry import register_class

from ..character_base import PhysicalNormalAttackBase
from .shenhe_3_7 import (
    Shenhe_3_7,
    SpringSpiritSummoning as SS_3_7,
    DivineMaidensDeliverance,
)


class SpringSpiritSummoning(SS_3_7):
    max_usage: int = 2


class Shenhe_4_2(Shenhe_3_7):
    version: Literal["4.2"] = "4.2"
    skills: List[
        PhysicalNormalAttackBase | SpringSpiritSummoning | DivineMaidensDeliverance
    ] = []

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Dawnstar Piercer",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SpringSpiritSummoning(),
            DivineMaidensDeliverance(),
        ]


register_class(Shenhe_4_2)



from typing import List, Literal

from ....utils.class_registry import register_class

from ...consts import DieColor

from ...struct import Cost

from .xiangling_3_8 import Crossfire_4_2, Xiangling_3_8 as X_3_8
from .xiangling_3_8 import Pyronado as P_3_8
from .xiangling_3_8 import DoughFu, GuobaAttack


class Pyronado(P_3_8):
    damage: int = 2


class Crossfire_3_3(Crossfire_4_2):
    version: Literal["3.3"]
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=4)


class Xiangling_3_3(X_3_8):
    version: Literal["3.3"]
    skills: List[DoughFu | GuobaAttack | Pyronado] = []

    def _init_skills(self) -> None:
        self.skills = [DoughFu(), GuobaAttack(), Pyronado()]


register_class(Xiangling_3_3 | Crossfire_3_3)



from typing import List, Literal

from ....utils.class_registry import register_class

from .xingqiu_4_1 import Xingqiu_4_1 as X_4_1
from .xingqiu_4_1 import Raincutter as R_4_1
from .xingqiu_4_1 import GuhuaStyle, FatalRainscreen


class Raincutter(R_4_1):
    name: Literal["Raincutter"] = "Raincutter"
    damage: int = 1


class Xingqiu_3_6(X_4_1):
    version: Literal["3.6"]
    skills: List[GuhuaStyle | FatalRainscreen | Raincutter] = []

    def _init_skills(self) -> None:
        self.skills = [GuhuaStyle(), FatalRainscreen(), Raincutter()]


register_class(Xingqiu_3_6)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import SkillEndEventArguments, SwitchCharacterEventArguments

from ...consts import (
    ELEMENT_TO_ATK_UP_ICON,
    ELEMENT_TO_DAMAGE_TYPE,
    CostLabels,
    DamageType,
    ElementType,
    IconType,
    ObjectPositionType,
    SkillType,
)

from ...action import CreateObjectAction, RemoveObjectAction

from ...modifiable_values import CostValue, DamageIncreaseValue

from .base import RoundTeamStatus, UsageTeamStatus


class Stormzone_3_7(UsageTeamStatus):
    name: Literal["Stormzone"] = "Stormzone"
    desc: Literal["", "talent"] = ""
    version: Literal["3.7"] = "3.7"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    talent_activated: bool = False
    decrease_cost_success: bool = False

    def renew(self, new_status: "Stormzone_3_7") -> None:
        super().renew(new_status)
        if new_status.talent_activated:
            self.talent_activated = True
            self.desc = "talent"

    def __init__(self, *argv, **kwargs) -> None:
        super().__init__(*argv, **kwargs)
        if self.talent_activated:
            # change desc
            self.desc = "talent"

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        Decrease sw_char cost.
        """
        if value.position.player_idx != self.position.player_idx:
            # not this player, do nothing
            return value
        if value.cost.label & CostLabels.SWITCH_CHARACTER.value == 0:
            # not sw_char, do nothing
            return value
        if value.cost.decrease_cost(None):
            # decrease success
            if mode == "REAL":
                self.decrease_cost_success = True
                self.usage -= 1
        return value

    def event_handler_SWITCH_CHARACTER(
        self, event: SwitchCharacterEventArguments, match: Any
    ) -> List[CreateObjectAction | RemoveObjectAction]:
        """
        if this status is talent activated, them create Winds of Harmony.
        Check if need to remove self.
        """
        if not self.decrease_cost_success:
            # not decrease cost success, do nothing
            return []
        self.decrease_cost_success = False
        ret: List[CreateObjectAction | RemoveObjectAction] = []
        if self.talent_activated:
            # create Winds of Harmony
            ret.append(
                CreateObjectAction(
                    object_name="Winds of Harmony",
                    object_position=self.position,
                    object_arguments={},
                )
            )
        ret += self.check_should_remove()
        return ret


# Usage status, will not disappear until usage is 0
class WindsOfHarmony_3_7(RoundTeamStatus):
    name: Literal["Winds of Harmony"] = "Winds of Harmony"
    version: Literal["3.7"] = "3.7"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode=Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        Decrease normal attack cost.
        """
        if self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # this character use skill, check if normal attack
            skill = match.get_object(value.position)
            skill_type: SkillType = skill.skill_type
            if skill_type == SkillType.NORMAL_ATTACK:
                # normal attack, decrease cost
                if value.cost.decrease_cost(None):
                    # decrease success
                    if mode == "REAL":
                        self.usage -= 1
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        return self.check_should_remove()


class PoeticsOfFuubutsu_3_8(UsageTeamStatus):
    name: Literal[
        "Poetics of Fuubutsu: Pyro",
        "Poetics of Fuubutsu: Hydro",
        "Poetics of Fuubutsu: Electro",
        "Poetics of Fuubutsu: Cryo",
    ]
    version: Literal["3.8"] = "3.8"
    usage: int = 2
    max_usage: int = 2
    element: ElementType = ElementType.NONE
    icon_type: Literal[
        IconType.ATK_UP_FIRE,
        IconType.ATK_UP_WATER,
        IconType.ATK_UP_ELEC,
        IconType.ATK_UP_ICE,
        IconType.ATK_UP,
    ] = IconType.ATK_UP

    def __init__(self, *argv, **kwargs) -> None:
        super().__init__(*argv, **kwargs)
        element_name = self.name.split()[-1]
        element = ElementType[element_name.upper()]
        self.element = element
        self.icon_type = ELEMENT_TO_ATK_UP_ICON[element]  # type: ignore

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        """
        If our character skill or summon deal corresponding elemental DMG,
        increase DMG.
        """
        if value.damage_type != DamageType.DAMAGE:
            # not damage, do nothing
            return value
        if value.position.player_idx != self.position.player_idx:
            # not this player, do nothing
            return value
        if value.damage_elemental_type != ELEMENT_TO_DAMAGE_TYPE[self.element]:
            # not corresponding elemental DMG, do nothing
            return value
        if value.position.area not in [
            ObjectPositionType.SKILL,
            ObjectPositionType.SUMMON,
        ]:
            # not character or summon, do nothing
            return value
        if value.damage_from_element_reaction:
            # damage from elemental reaction, do nothing
            return value
        # increase DMG
        assert mode == "REAL"
        self.usage -= 1
        value.damage += 1
        return value


register_class(Stormzone_3_7 | WindsOfHarmony_3_7 | PoeticsOfFuubutsu_3_8)



from typing import Dict, List, Literal
from lpsim.server.action import Actions

from lpsim.server.card.equipment.artifact.base import ArtifactBase
from lpsim.server.event import RoundEndEventArguments
from lpsim.server.match import Match
from lpsim.server.modifiable_values import CostValue
from lpsim.server.struct import Cost
from lpsim.utils.class_registry import register_class
from lpsim.utils.desc_registry import DescDictType


class GoldenTroupesReward_4_5(ArtifactBase):
    name: Literal["Golden Troupe's Reward"] = "Golden Troupe's Reward"
    version: Literal["4.5"] = "4.5"
    cost: Cost = Cost()
    usage: int = 0
    max_usage: int = 2

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Match
    ) -> List[Actions]:
        if self.position.satisfy("source area=character active=false", match=match):
            self.usage = min(self.usage + 1, self.max_usage)
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Match, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        If self use any skill, or equip talent, cost decrease based on self usage.
        """
        if not self._check_value_self_skill_or_talent(value, match):
            return value
        # decrease cost
        for _ in range(self.usage):
            if value.cost.decrease_cost(value.cost.elemental_dice_color):
                if mode == "REAL":
                    self.usage -= 1
        return value


desc: Dict[str, DescDictType] = {
    "ARTIFACT/Golden Troupe's Reward": {
        "names": {"en-US": "Golden Troupe's Reward", "zh-CN": "黄金剧团的奖赏"},
        "descs": {
            "4.5": {
                "en-US": 'End Phase: If the character to which this is attached is on standby, this card gains 1 "Recompense" point. (Max 2 points)\nWhen you play a Talent card or a Character uses an Elemental Skill: For each "Recompense" point this card has, consume it to spend 1 less Elemental Die.\n(A character can equip a maximum of 1 Artifact)',  # noqa: E501
                "zh-CN": "结束阶段：如果所附属角色在后台，则此牌累积1点「报酬」。（最多累积2点）\n对角色打出「天赋」或角色使用「元素战技」时：此牌每有1点「报酬」，就将其消耗，以少花费1个元素骰。\n（角色最多装备1件「圣遗物」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Artifact_HuangjinjutuanXiaojian.png",  # noqa: E501
        "id": 312025,
    },
}


register_class(GoldenTroupesReward_4_5, desc)



import os
import platform
import uvicorn
import signal
import time
import logging
import random
import threading
from fastapi.responses import JSONResponse
from typing import List, Tuple
from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from starlette.middleware.gzip import GZipMiddleware
from multiprocessing import Process, Queue

from .http_server import HTTPServer
from .__version__ import __version_tuple__, __version__, __frontend_version__


class OneRoomWorker(Process):
    def __init__(self, cmd_queue: Queue, resp_queue: Queue, post_queue: Queue):
        """
        Initializes the OneRoomWorker class.

        Args:
            cmd_queue (Queue): The queue used to send arguments and messages about
                running the server.
            resp_queue (Queue): The queue used to send responses.
            post_queue (Queue): The queue used to send timestamps when a POST request
                is received.
        """
        super().__init__()
        self.cmd_queue = cmd_queue
        self.resp_queue = resp_queue
        self.post_queue = post_queue

    def run(self):
        """
        Receives init_args, run_args and room_port_range from `self.cmd_queue`,
        creates an `HTTPServer` and runs it with the given args until a SIGINT is
        received, then sends a message to the queue and waits for the next args. If
        None is received, it exits.

        Raises:
            Exception: Describes any exceptions that might be thrown.
        """
        while True:
            try:
                init_args, run_args, room_port_range = self.cmd_queue.get()
                if init_args is None:
                    # when receive None, return
                    break
                self.server = HTTPServer(**init_args)
                self.port = None

                # self.server.app.add_middleware(
                #     PostTimestampMiddleware, post_queue = self.post_queue)

                @self.server.app.middleware("http")
                async def post_timestamp_sender(request: Request, call_next):
                    if request.method == "POST":
                        self.post_queue.put(time.time())
                    response = await call_next(request)
                    return response

                @self.server.app.on_event("startup")
                def startup_event():
                    self.resp_queue.put(str(self.port))

                max_retry = 100
                for _ in range(max_retry):
                    try:
                        # try to occupy random port from room_port_range
                        self.port = random.randint(*room_port_range)
                        self.server.run(**run_args, port=self.port)
                        self.server.save_log(self.server.reset_log_save_file)
                        # when receive sigint, gracefully end server and send
                        # message to queue
                        self.resp_queue.put(
                            {
                                "decks": self.server.uploaded_deck_codes,
                            }
                        )
                        break
                    except SystemExit:
                        # start failed, send message to queue
                        self.resp_queue.put("invalid and retry")
                else:
                    self.resp_queue.put("server failed to start")
            except KeyboardInterrupt:
                # when receive sigint when waiting for args, simply ignore it
                pass


class HTTPRoomServer:
    """
    HTTP room server that host multiple rooms, each room is a HTTP server.
    Room will occupy a port, room server will deal with requests from client,
    to reset/start a new room etc, and tell client the room name and
    link of the server.
    It only starts rooms with same rules now. When running on Linux, it will
    collect deck codes of stopped rooms, but not on Windows.

    Args:
        max_rooms (int): Max number of rooms that can be created. All rooms
            will be created at the beginning.
        room_port_range (List[int]): Range of ports that can be used to
            create rooms. ports will be chosen in random order.  If no port is
            available but the room number is not reached, the room server will
            raise error and exit.
        room_timeout (int): Timeout of a room in seconds. If any valid action
            is performed, the timeout will be reset. Room server will
            periodically check the timeout and close the room if timeout.
        admin_password (str): Password for admin. Admin can reset or remove
            any room. If configured, only admin can see all room ids, and
            decks used for different rooms.
        allow_origins (List[str]): List of origins that are allowed to connect.
        reset_log_save_path (str): Path to save reset log.
        init_args (dict): Args for HTTPServer.__init__, e.g. match_config.
        run_args (dict): Args for HTTPServer.run, e.g. HTTPS certificate. But
            not port, which will be chosen randomly from room_port_range.
    """

    def __init__(
        self,
        max_rooms: int = 2,
        port: int = 7999,
        room_port_range: List[int] = [8000, 9000],
        room_timeout: int = 3600,
        admin_password: str = "",
        allow_origins: List[str] = ["*"],
        init_args: dict = {},
        run_args: dict = {},
    ):
        self.max_rooms = max_rooms
        self.port = port
        self.room_port_range = room_port_range
        self.room_timeout = room_timeout
        self.admin_password = admin_password
        self.allow_origins = allow_origins
        self.init_args = init_args
        self.run_args = run_args

        self.deck_history: List[Tuple[str, str]] = []

        self._check_time_interval = 5

        if platform.system() == "Windows":
            logging.warning(
                "Windows system detected, room server will use terminate "
                "instead of graceful shutdown, and cannot collect uploaded "
                "deck information."
            )

        self.app = FastAPI()
        app = self.app

        app.add_middleware(
            CORSMiddleware,
            allow_origins=allow_origins,
            # allow_origins=['*'],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )
        app.add_middleware(GZipMiddleware)

        @app.get("/version")
        async def get_version():
            """
            Return the version of lpsim.

            Return:
                version: version string
                version_tuple: version tuple
                support_version: support version of frontend
                info: information of the server
            """
            return {
                "version": __version__,
                "version_tuple": __version_tuple__,
                "support_version": __frontend_version__,
                "info": {
                    "class": "HTTPServer",
                },
            }

        @app.post("/room/{room_name}")
        def post_room_name(room_name: str):
            """
            Create a new room with given name. If room is already created,
            return its port.

            Args:
                room_name (str): The name of the room.

            Returns:
                JSONResponse: The response containing the port (if success) and status
                of the room. When status is exist or created, the corresponding port
                will be returned. Otherwise, only status will be returned.
            """
            if room_name in self.room_names:
                idx = self.room_names.index(room_name)
                return JSONResponse({"port": self.room_ports[idx], "status": "exist"})
            else:
                if None not in self.room_names:
                    return JSONResponse({"status": "full"})
                # create new one
                idx = self.room_names.index(None)
                init_args = self.init_args.copy()
                init_args["room_name"] = room_name
                cmd_q, resp_q, post_q = self.queues[idx]
                cmd_q.put((init_args, self.run_args, self.room_port_range))
                while True:
                    resp = resp_q.get()
                    if resp == "server failed to start":
                        return JSONResponse({"status": "failed"})
                    else:
                        # is port, but may start failed, need to check
                        # whether fail message is sent
                        if resp_q.empty():
                            # if is empty, wait a minute and check again
                            time.sleep(0.1)
                        if resp_q.empty():
                            # queue empty, start success
                            self.room_names[idx] = room_name
                            self.room_ports[idx] = int(resp)
                            self.room_active_times[idx] = time.time()
                            return JSONResponse(
                                {"port": self.room_ports[idx], "status": "created"}
                            )
                        else:
                            # start failed with message
                            resp = resp_q.get()
                            assert resp == "invalid and retry"

        @app.delete("/room/{room_name}")
        def delete_room_name(room_name: str, password: str = ""):
            """
            Delete a room with given name. If room is not created, return
            error.

            Args:
                room_name (str): The name of the room.
                password (str): The password for the admin.

            Returns:
                JSONResponse: The response containing the action status (not exist or
                deleted).
            """
            if password != self.admin_password:
                return JSONResponse(status_code=403, content="wrong password")
            if room_name not in self.room_names:
                return JSONResponse({"status": "not exist"}, 404)
            else:
                idx = self.room_names.index(room_name)
                self._delete_one_room(idx)
                return JSONResponse({"status": "deleted"})

        @app.get("/rooms")
        def get_rooms(password: str = ""):
            """
            Get all rooms' name and port.

            Args:
                password (str): The password for the admin.

            Returns:
                JSONResponse: The response containing the rooms' name, port and timeout.
            """
            if password != self.admin_password:
                return JSONResponse(status_code=403, content="wrong password")
            return JSONResponse(
                {
                    "rooms": [
                        {
                            "name": name,
                            "port": port,
                            "timeout": int(self.room_timeout - time.time() + at),
                        }
                        for name, port, at in zip(
                            self.room_names, self.room_ports, self.room_active_times
                        )
                        if name is not None
                    ]
                }
            )

        @app.get("/decks")
        def get_decks(password: str = ""):
            """
            Get all decks used in rooms.

            Args:
                - password (str): The password for the admin.

            Returns:
                - JSONResponse: The response containing the decks used in rooms.
            """
            if password != self.admin_password:
                return JSONResponse(status_code=403, content="wrong password")
            return JSONResponse(self.deck_history)

    def _create_room_workers(self):
        max_rooms = self.max_rooms
        self.workers = []
        self.queues: List[Tuple[Queue, Queue, Queue]] = []
        # when name is None, room has no name and is ended
        self.room_names: List[str | None] = [None] * max_rooms
        self.room_ports: List[int] = [0] * max_rooms
        self.room_active_times: List[float] = [1e100] * max_rooms
        for _ in range(max_rooms):
            cmd_queue = Queue()
            resp_queue = Queue()
            post_queue = Queue()
            worker = OneRoomWorker(cmd_queue, resp_queue, post_queue)
            worker.start()
            self.workers.append(worker)
            self.queues.append((cmd_queue, resp_queue, post_queue))

    def run(self, *argv, **kwargs):
        """
        Create room workers, start them and run room server.
        When HTTPRoomServer stopped, stop all workers and stop check interval.
        All args will be passed to uvicorn.run, except port, which is not allowed.
        """
        if len(argv):
            raise ValueError("positional arguments not supported")

        # create room workers
        self._create_room_workers()

        # start check interval
        self._stop_interval = False
        self._stop_interval_success = False
        self._check_interval_func()

        # start room server
        assert "port" not in kwargs
        kwargs = kwargs.copy()
        kwargs["port"] = self.port
        uvicorn.run(self.app, **kwargs)

        # after run, mark _stop_interval as True, and wait for stop success
        logging.warning(
            "stopping interval check, it should take at most "
            f"{self._check_time_interval} seconds."
        )
        self._stop_interval = True
        while not self._stop_interval_success:
            time.sleep(0.1)
        self._stop_interval = False
        self._stop_interval_success = False

        # after run, stop all worker rooms
        for idx, (worker, queue, name) in enumerate(
            zip(self.workers, self.queues, self.room_names)
        ):
            if name is not None:
                self._delete_one_room(idx)

        if platform.system() != "Windows":
            # for OS other than windows, send stop signal to all workers and
            # join
            for worker, queue in zip(self.workers, self.queues):
                queue[0].put((None, None, None))
                worker.join()

    def _delete_one_room(self, room_idx: int):
        if platform.system() == "Windows":
            self._delete_one_room_by_terminate(room_idx)
        else:
            self._delete_one_room_gracefully(room_idx)

    def _delete_one_room_gracefully(self, room_idx: int):
        room_name = self.room_names[room_idx]
        assert room_name is not None
        worker = self.workers[room_idx]
        pid = worker.pid
        os.kill(pid, signal.SIGINT)
        msg = self.queues[room_idx][1].get()
        decks = msg["decks"]
        for deck in decks:
            self.deck_history.append((room_name, deck))
        room_port = self.room_ports[room_idx]
        logging.warning(f"room {room_name}:{room_idx}:{room_port} stopped")
        self.room_names[room_idx] = None
        self.room_ports[room_idx] = 0

    def _delete_one_room_by_terminate(self, room_idx: int):
        room_name = self.room_names[room_idx]
        assert room_name is not None
        worker = self.workers[room_idx]
        pid = worker.pid
        os.kill(pid, signal.SIGTERM)
        room_port = self.room_ports[room_idx]
        logging.warning(f"room {room_name}:{room_idx}:{room_port} terminated")
        # as it terminates, we cannot receive deck histories, and we need to
        # re-create a new worker.
        new_worker = OneRoomWorker(*self.queues[room_idx])
        new_worker.start()
        self.workers[room_idx] = new_worker
        self.room_names[room_idx] = None
        self.room_ports[room_idx] = 0

    def _remove_unused_rooms(self):
        """
        Check if any room is not used for a long time, and remove it if so.
        """
        for room_idx, room_name in enumerate(self.room_names):
            if room_name is None:
                continue
            post_queue = self.queues[room_idx][2]
            while not post_queue.empty():
                self.room_active_times[room_idx] = post_queue.get()
            if time.time() - self.room_active_times[room_idx] > self.room_timeout:
                logging.warning(f"room {room_name}:{room_idx} timeout")
                self._delete_one_room(room_idx)

    def _check_interval_func(self):
        self._remove_unused_rooms()
        if self._stop_interval:
            self._stop_interval_success = True
            return
        threading.Timer(self._check_time_interval, self._check_interval_func).start()


__all__ = ["HTTPRoomServer"]



from typing import List, Literal

from ....utils.class_registry import register_class
from .arataki_itto_3_6 import (
    AratakiItto_3_6,
    FightClubLegend,
    MasatsuZetsugiAkaushiBurst,
    RoyalDescentBeholdIttoTheEvil as RDBITE_3_6,
)


class RoyalDescentBeholdIttoTheEvil(RDBITE_3_6):
    damage: int = 4
    version: Literal["4.2"] = "4.2"


class AratakiItto_4_2(AratakiItto_3_6):
    version: Literal["4.2"] = "4.2"
    skills: List[
        FightClubLegend | MasatsuZetsugiAkaushiBurst | RoyalDescentBeholdIttoTheEvil
    ] = []

    def _init_skills(self) -> None:
        self.skills = [
            FightClubLegend(),
            MasatsuZetsugiAkaushiBurst(),
            RoyalDescentBeholdIttoTheEvil(),
        ]


register_class(AratakiItto_4_2)



from typing import Dict, List, Literal

from pydantic import PrivateAttr
from lpsim.server.action import Actions, CreateObjectAction, RemoveObjectAction
from lpsim.server.card.support.locations import LocationBase
from lpsim.server.consts import DamageType, IconType, ObjectPositionType, ObjectType
from lpsim.server.event import (
    ReceiveDamageEventArguments,
    RoundPrepareEventArguments,
    UseCardEventArguments,
)
from lpsim.server.match import Match
from lpsim.server.modifiable_values import UseCardValue
from lpsim.server.status.team_status.base import RoundTeamStatus
from lpsim.server.struct import Cost, ObjectPosition
from lpsim.utils.class_registry import register_class
from lpsim.utils.desc_registry import DescDictType


class FortressOfMeropideStatus_4_5(RoundTeamStatus):
    name: Literal["Fortress of Meropide"] = "Fortress of Meropide"
    version: Literal["4.5"] = "4.5"
    usage: int = 1
    max_usage: int = 1
    icon_type: IconType = IconType.DEBUFF

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Match
    ) -> List[RemoveObjectAction]:
        """
        check should remove
        """
        return self.check_should_remove()

    def value_modifier_USE_CARD(
        self, value: UseCardValue, match: Match, mode: Literal["TEST", "REAL"]
    ) -> UseCardValue:
        """
        When self use a event card, make using card failed.
        """
        if value.position.player_idx != self.position.player_idx:
            # not self, return
            return value
        if value.card.type in [ObjectType.CARD, ObjectType.ARCANE]:
            # is target card type
            assert self.usage > 0
            assert mode == "REAL"
            if value.use_card:  # pragma: no branch
                value.use_card = False
                self.usage -= 1
        return value


class FortressOfMeropide_4_5(LocationBase):
    name: Literal["Fortress of Meropide"] = "Fortress of Meropide"
    version: Literal["4.5"] = "4.5"
    cost: Cost = Cost(same_dice_number=1)
    usage: int = 0
    _max_usage: int = PrivateAttr(4)
    icon_type: IconType = IconType.COUNTER

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[Actions]:
        """
        When self active receive damage or heal, get 1 counter
        """
        if event.final_damage.damage_type not in [DamageType.DAMAGE, DamageType.HEAL]:
            return []
        if self.position.not_satisfy(
            "both player=same and source area=support and target active=true",
            target=event.final_damage.target_position,
            match=match,
        ):
            return []
        self.usage = min(self._max_usage, self.usage + 1)
        return []

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        """
        When usage reach max, and status not exist, create status
        """
        if self.usage < self._max_usage:
            return []
        target = self.query_one(match, f'opponent team_status "name={self.name}"')
        if target is not None:
            # status already exist
            return []
        self.usage = 0
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=1 - self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=0,
                ),
                object_arguments={},
            )
        ]


desc: Dict[str, DescDictType] = {
    "TEAM_STATUS/Fortress of Meropide": {
        "names": {"en-US": "Fortress of Meropide", "zh-CN": "梅洛彼得堡"},
        "descs": {
            "4.5": {
                "en-US": "1 Event Card played by you this Round will have no effect.",
                "zh-CN": "本回合中我方打出的1张事件牌无效",
            }
        },
    },
    "SUPPORT/Fortress of Meropide": {
        "names": {"en-US": "Fortress of Meropide", "zh-CN": "梅洛彼得堡"},
        "descs": {
            "4.5": {
                "en-US": 'When your active character takes DMG or is healed: This card gains 1 "Prohibition" point. (Maximum 4 points)\nWhen Action Phase begins: If this card has accumulated 4 "Prohibition" points, then 4 points will be consumed and 1 Strictly Prohibited will be created in your opponent\'s playing field. (1 Event Card played by your opponent this Round will have no effect.)',  # noqa: E501
                "zh-CN": "我方出战角色受到伤害或治疗后：此牌累积1点「禁令」。（最多累积到4点）\n行动阶段开始时：如果此牌已有4点「禁令」，则消耗4点，在对方场上生成严格禁令。（本回合中打出的1张事件牌无效）",  # noqa: E501
            }
        },
        "image_path": "cardface/Assist_Location_BideBao.png",  # noqa: E501
        "id": 321018,
    },
}


register_class(FortressOfMeropide_4_5 | FortressOfMeropideStatus_4_5, desc)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions

from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Skills


class GuhuaStyle(PhysicalNormalAttackBase):
    name: Literal["Guhua Style"] = "Guhua Style"
    cost: Cost = PhysicalNormalAttackBase.get_cost(ElementType.HYDRO)


class FatalRainscreen(ElementalSkillBase):
    name: Literal["Fatal Rainscreen"] = "Fatal Rainscreen"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack, application and create object
        """
        ret = [
            self.attack_opposite_active(match, self.damage, self.damage_type),
            self.charge_self(1),
        ]
        ele_app = self.element_application_self(match, DamageElementalType.HYDRO)
        ret[0].damage_value_list += ele_app.damage_value_list
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        talent = character.talent
        if talent is not None:
            ret.append(
                self.create_team_status(
                    "Rain Sword",
                    {
                        "version": talent.version,
                        "usage": 3,
                        "max_usage": 3,
                    },
                )
            )
        else:
            ret.append(self.create_team_status("Rain Sword", {"version": "3.3"}))
        return ret


class Raincutter(ElementalBurstBase):
    name: Literal["Raincutter"] = "Raincutter"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack, application and create object
        """
        ret = [
            self.charge_self(-self.cost.charge),
            self.attack_opposite_active(match, self.damage, self.damage_type),
        ]
        ele_app = self.element_application_self(match, DamageElementalType.HYDRO)
        ret[1].damage_value_list += ele_app.damage_value_list
        ret.append(self.create_team_status("Rainbow Bladework"))
        return ret


# Talents


class TheScentRemained_3_3(SkillTalent):
    name: Literal["The Scent Remained"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Xingqiu"] = "Xingqiu"
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=4)
    skill: Literal["Fatal Rainscreen"] = "Fatal Rainscreen"


class TheScentRemained_4_2(TheScentRemained_3_3):
    name: Literal["The Scent Remained"]
    version: Literal["4.2"] = "4.2"
    character_name: Literal["Xingqiu"] = "Xingqiu"
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3)
    skill: Literal["Fatal Rainscreen"] = "Fatal Rainscreen"


# character base


class Xingqiu_4_1(CharacterBase):
    name: Literal["Xingqiu"]
    version: Literal["4.1"] = "4.1"
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[GuhuaStyle | FatalRainscreen | Raincutter] = []
    faction: List[FactionType] = [FactionType.LIYUE]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            GuhuaStyle(),
            FatalRainscreen(),
            Raincutter(),
        ]


register_class(Xingqiu_4_1 | TheScentRemained_3_3 | TheScentRemained_4_2)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import DefendSummonBase

from ...modifiable_values import DamageValue

from ...action import Actions, MakeDamageAction, RemoveObjectAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    AOESkillBase,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class BaronBunny_3_7(DefendSummonBase):
    name: Literal["Baron Bunny"] = "Baron Bunny"
    version: Literal["3.7"] = "3.7"
    usage: int = 1
    max_usage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    damage: int = 2
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 2
    attack_until_run_out_of_usage: bool = True


# Skills


class Sharpshooter(PhysicalNormalAttackBase):
    name: Literal["Sharpshooter"] = "Sharpshooter"
    cost: Cost = PhysicalNormalAttackBase.get_cost(ElementType.PYRO)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If has talent, and Baron Bunny in area, do extra 3 pyro damage and
        remove Baron Bunny.
        """
        ret = super().get_actions(match)
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        talent = character.talent
        if talent is not None:
            summons = match.player_tables[self.position.player_idx].summons
            for summon in summons:
                if summon.name == "Baron Bunny":
                    target = match.player_tables[
                        1 - self.position.player_idx
                    ].get_active_character()
                    ret.append(
                        MakeDamageAction(
                            damage_value_list=[
                                DamageValue(
                                    position=summon.position,
                                    damage_type=DamageType.DAMAGE,
                                    target_position=target.position,
                                    damage=talent.damage,
                                    damage_elemental_type=DamageElementalType.PYRO,
                                    cost=Cost(),
                                )
                            ]
                        )
                    )
                    ret.append(RemoveObjectAction(object_position=summon.position))
                    return ret
        return ret


class ExplosivePuppet(ElementalSkillBase):
    name: Literal["Explosive Puppet"] = "Explosive Puppet"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object only
        """
        return [
            self.charge_self(1),
            self.create_summon("Baron Bunny"),
        ]


class FieryRain(ElementalBurstBase, AOESkillBase):
    name: Literal["Fiery Rain"] = "Fiery Rain"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.PYRO
    back_damage: int = 2
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=3, charge=2
    )


# Talents


class BunnyTriggered_4_2(SkillTalent):
    name: Literal["Bunny Triggered"]
    version: Literal["4.2"] = "4.2"
    character_name: Literal["Amber"] = "Amber"
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=3)
    skill: Literal["Explosive Puppet"] = "Explosive Puppet"
    damage: int = 4


class BunnyTriggered_3_7(SkillTalent):
    name: Literal["Bunny Triggered"]
    version: Literal["3.7"] = "3.7"
    character_name: Literal["Amber"] = "Amber"
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=3)
    skill: Literal["Explosive Puppet"] = "Explosive Puppet"
    damage: int = 3


# character base


class Amber_3_7(CharacterBase):
    name: Literal["Amber"]
    version: Literal["3.7"] = "3.7"
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[Sharpshooter | ExplosivePuppet | FieryRain] = []
    faction: List[FactionType] = [FactionType.MONDSTADT]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [Sharpshooter(), ExplosivePuppet(), FieryRain()]


register_class(Amber_3_7 | BunnyTriggered_3_7 | BunnyTriggered_4_2 | BaronBunny_3_7)
